/* *****************************************************
 * Pixelshuffler is a command-line tool to optimize a Koala image file in order to achieve a smaller compressed file
 *
 * September-2020 V0.4
 * Wilfried Elmenreich
 * Code under CC-BY-4.0 license
 * **************************************************** */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>


typedef unsigned char int8_t;

typedef struct { 
 int8_t startlo;
 int8_t starthi;
 int8_t data[0x10000];
 int Bstart;
 int Sstart;
 int Cstart;
 int Gstart; 
 int Bend;
 int Send;
 int Cend;
 int length;
} cbmfile_t;

int VERBOSE=1;
char *progname;
char *palette="17df3a5ce842b960";
char *fstructure="BSCG";

cbmfile_t load_file(char* filename) {
  FILE *fp;
  cbmfile_t filedata;
  fp = fopen(filename, "rb");
  filedata.startlo = fgetc(fp);
  filedata.starthi = fgetc(fp);
  filedata.length=0;
  while(1) {
    int c=fgetc(fp);
    if (feof(fp)) break;  
    filedata.data[filedata.length++]=c;
  }
  fclose(fp);

  if (VERBOSE) {
      int start=filedata.startlo+256*filedata.starthi;
      int end=start+filedata.length;
      printf("Read file '%s' $%04x-$%04x\n",filename,start,end);
   }
  if (filedata.length!=10001) //if the size fits, we assume this is a Koala image
	  printf("Warning! File length is not 10001 bytes but %i.\n",filedata.length);
  filedata.Bstart=0;
  filedata.Bend=filedata.Sstart=8000;
  filedata.Send=filedata.Cstart=9000;
  filedata.Cend=filedata.Gstart=10000;  
  
  filedata.data[10000]= filedata.data[10000] & 0x0f;
return filedata;
}
 
void save_file(cbmfile_t filedata, char* filename) {
   FILE *fp;
   fp = fopen(filename, "wb");
   fprintf(fp, "%c", filedata.startlo);
   fprintf(fp, "%c", filedata.starthi);
   for (int i=0; i<filedata.length; i++)
      fprintf(fp, "%c", filedata.data[i]);
   fclose(fp);
   if (VERBOSE) {
      int start=filedata.startlo+256*filedata.starthi;
      int end=start+filedata.length;
      printf("Written file '%s' $%04x-$%04x\n",filename,start,end);
   }
}

cbmfile_t shuffle_nibbles(cbmfile_t filedata) {
	for (int i=0;i<8000;i+=2) {
		int8_t a=filedata.data[i];
		int8_t b=filedata.data[i+1];
		filedata.data[i]=(a & 0x0f) + ((b & 0x0f)<<4);
		filedata.data[i+1]=(( a & 0xf0)>>4) + (b & 0xf0);	
	}
	return filedata;
}

cbmfile_t shuffle_bytes0213(cbmfile_t filedata) {
	for (int i=0;i<8000;i+=4) {
		//int8_t a=filedata.data[i];
		int8_t b=filedata.data[i+1];
		int8_t c=filedata.data[i+2];
		//int8_t d=filedata.data[i+3];		
		filedata.data[i+1]=c;
		filedata.data[i+2]=b;	
	}
	return filedata;
}

cbmfile_t shuffle_bytes02461357(cbmfile_t filedata) {
	for (int i=0;i<8000;i+=8) {
		//int8_t a=filedata.data[i];
		int8_t b=filedata.data[i+1];
		int8_t c=filedata.data[i+2];
		int8_t d=filedata.data[i+3];
		int8_t e=filedata.data[i+4];	
		int8_t f=filedata.data[i+5];
		int8_t g=filedata.data[i+6];	
		//int8_t h=filedata.data[i+7];		
		filedata.data[i+1]=c;
		filedata.data[i+2]=e;	
		filedata.data[i+3]=g;
		filedata.data[i+4]=b;
		filedata.data[i+5]=d;
		filedata.data[i+6]=f;
	}
	return filedata;
}

cbmfile_t storediffs(cbmfile_t filedata,int length) {
	int8_t B=0;
	for (int i=0;i<length;i++) {
		filedata.data[i] = filedata.data[i]-B;
		B+=filedata.data[i];
	}
	return filedata;
}

cbmfile_t storexors(cbmfile_t filedata,int length) {
	int8_t B=0;
	for (int i=0;i<length;i++) {
		filedata.data[i] = filedata.data[i]^B;
		B^=filedata.data[i];
	}
	return filedata;
}

cbmfile_t reorganizecolors(cbmfile_t filedata) {
	int colorgradients[16]={16,16,16,16, 16,16,16,16, 16,16,16,16, 16,16,16,16};
	int bgcolor=filedata.data[10000];
	int blanked=0;
	
	//convert palette string to color numbers in colorgradients
	for(int i=0;i<16;i++) {
		int c=palette[i];
		if (c>='0' && c<='9')
			colorgradients[i]=c-'0';
		else
			colorgradients[i]=c-'a'+10;
	}

	//check if each color was given
	for(int i=0;i<16;i++) {
		if (colorgradients[i]==16) {
			fprintf(stderr,"Palette is missing a color!");
			exit(1);
		}
	}	
	
	//go through each of the 1000 4x8 blocks
	for (int c=0;c<1000;c++) {
		int cols[4]={0,0,0,0};
		int colors[16]={0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};
		int matrix[8][4];
		int colindex=0;
		for(int i=c*8;i<c*8+8;i++) {
			int b=filedata.data[i];
			//for each fat pixel
			for(int k=0;k<4;k++) {
				int bits=(b >> (k+k)) & 0x03;
				int color=-1;
				cols[bits]++;
				switch(bits) {
				case 0:
					color=bgcolor;
					break;
				case 1:
					color=(filedata.data[8000+c] & 0xf0)>>4;
					break;	
				case 2:
					color=filedata.data[8000+c] & 0x0f;
					break;		
				case 3:
					color=filedata.data[9000+c] & 0x0f;
					break;												
				}
				colors[color]++;	
				matrix[i-c*8][k]=color; // we will later rebuild the 4x8 block based on this
			}
	    }
	    colors[bgcolor]=0; // we don't care how many background pixels are here
	    colindex=1;
	    
	    for(int i=15; i>=0; i--) {
	    	if (colors[colorgradients[i]]==0) continue;	
	    	colors[colorgradients[i]]=colindex;
	    	switch(colindex) {
			case 1:
				filedata.data[8000+c] = colorgradients[i] << 4;
				break;	
			case 2:
				filedata.data[8000+c] |= colorgradients[i];
				break;		
			case 3:
				filedata.data[9000+c] = colorgradients[i];
				break;												
			}
			colindex++;
		}
		for(int i=colindex; i<4; i++) {
	    	switch(colindex) {
			case 1:
				blanked++;				
				filedata.data[8000+c] = (filedata.data[8000+c] & 0x0f) + (filedata.data[8000+c-1] & 0xf0);
				break;	
			case 2:
				blanked++;
				filedata.data[8000+c] = (filedata.data[8000+c] & 0xf0) + (filedata.data[8000+c-1] & 0x0f);
				break;		
			case 3:
				blanked++;
				filedata.data[9000+c] = filedata.data[9000+c-1];
				break;												
			}	
		}
		for(int i=c*8;i<c*8+8;i++) {
			int bytepixels=0;
			for(int k=0;k<4;k++) {
				bytepixels+=colors[matrix[i-c*8][k]] << (k+k);
			}
			filedata.data[i]=bytepixels;
		}
	}
    if (VERBOSE) {
      printf("Eliminated %i color entries.\n",blanked);
    }	
	return filedata;
}

cbmfile_t blankoutunusedcolors(cbmfile_t filedata) {
	//ununsed colors are replaced by the same color in the previous pixel
	int blanked=0;
	for (int c=0;c<1000;c++) {
		int cols[4]={0,0,0,0};
		for(int i=c*8;i<c*8+8;i++) {
			int b=filedata.data[i];
			for(int k=0;k<4;k++) {
				cols[(b >> (k+k)) & 0x03]++;
			}
	    }
		if (cols[1]==0) {
			if ((filedata.data[8000+c] & 0xf0)>0) blanked++;			
			filedata.data[8000+c] = (filedata.data[8000+c] & 0x0f) + (filedata.data[8000+c-1] & 0xf0);
		}
		if (cols[2]==0) {
			if ((filedata.data[8000+c] & 0x0f)>0) blanked++;		
			filedata.data[8000+c] = (filedata.data[8000+c] & 0xf0) + (filedata.data[8000+c-1] & 0x0f);			
			//filedata.data[8000+c] &= 0xf0;			
		}
		if (cols[3]==0) {
			if ((filedata.data[9000+c] & 0x0f)!=(filedata.data[9000+c-1] & 0x0f)) blanked++;			
			filedata.data[9000+c] = filedata.data[9000+c-1];				
		}
	}
    if (VERBOSE) {
      printf("Eliminated %i color entries.\n",blanked);
    }	
	return filedata;
}


cbmfile_t packcolbytes(cbmfile_t filedata) {
	for (int i=0;i<500;i++) {
		filedata.data[9000+i] = (filedata.data[9000+i+i]&0x0f) + ((filedata.data[9000+i+i+1]&0x0f)<<4);
	}
	filedata.data[9500]=filedata.data[10000];
	filedata.Cend=filedata.Gstart=filedata.length=9500;
	return filedata;
}

cbmfile_t dintcolordata(cbmfile_t filedata) { //deinterleave color data, have 1000 halfbytes of color 01 then 1000 halfbytes of color 10
	int8_t dintdata[1000];
	for (int i=0;i<500;i++) {
	   dintdata[i]=(filedata.data[8000+i+i]&0x0f) + ((filedata.data[8001+i+i]&0x0f)<<4);
	   dintdata[500+i]=((filedata.data[8000+i+i]&0xf0)>>4) + (filedata.data[8001+i+i]&0xf0);
	}
	for (int i=0;i<1000;i++) {
	   filedata.data[8000+i] = dintdata[i];
	}	
	return filedata;
}

cbmfile_t columnwise(cbmfile_t filedata) {
	int8_t bitmap[8000];
	
	for (int col=0; col<40; col++)
	   for(int line=0;line<200;line++)
	      bitmap[col*200+line]=filedata.data[320*(line>>3)+col*8+(line & 7)];
	for (int i=0;i<8000;i++) {
		filedata.data[i] = bitmap[i];
	}
	return filedata;
}



cbmfile_t filestructure(cbmfile_t filedata, char* fstructure) {
	cbmfile_t filedata2;
	filedata2.length=0;
	
	filedata2.startlo = 0x00;
    filedata2.starthi = 0x60;
	while (*fstructure!='\0') {
		int8_t padbyte=0;
		
		printf("-%c-\n",*fstructure);
		switch(*fstructure) {
		case 'B':
			for (int i=filedata.Bstart; i<filedata.Bend; i++) {
				filedata2.Bstart=filedata2.length;
				filedata2.data[filedata2.length++]=filedata.data[i];
				filedata2.Bend=filedata2.length;			
			}
			break;
		case 'S':
			for (int i=filedata.Sstart; i<filedata.Send; i++) {
				filedata2.Sstart=filedata2.length;
				filedata2.data[filedata2.length++]=filedata.data[i];
				filedata2.Send=filedata2.length;			
			}		
			break;
		case 'C':
			for (int i=filedata.Cstart; i<filedata.Cend; i++) {
				filedata2.Cstart=filedata2.length;
				filedata2.data[filedata2.length++]=filedata.data[i];
				filedata2.Cend=filedata2.length;			
			}		
			break;	
		case 'G':
			filedata2.Gstart=filedata2.length;
			filedata2.data[filedata2.length++]=filedata.data[filedata.Gstart];
			break;			
		case 'p':
			printf("padding..\n");
			if (filedata2.length>0)
			   padbyte=filedata2.data[filedata2.length-1];
			for (int i=filedata2.length; i<1024 * ((filedata2.length+1023)/1024);i++)
			   filedata2.data[filedata2.length++]=padbyte;
			break;			
		}
		fstructure++;
	}
	return filedata2;
}


void usage(char *s) {
   if (s!=NULL) fprintf(stderr,"%s\n\n",s);
	
   fprintf(stderr,"Usage: %s [-i <infilename>] [-b] [-z] [-p palette] [-123458d] [-r] [-s structure] [-o <outfilename>]\n",progname);

   fprintf(stderr,"\nThe following options require the infile to be a Koala or scol64 image:\n");  
   fprintf(stderr,"\nOptions preserving koala format:\n");  
   fprintf(stderr," -b blank out unused colors\n");    
   fprintf(stderr," -z reorganize colors\n");       
   fprintf(stderr," -p ordered palette for reorganize colors, default is 17df3a5ce842b960\n");
      
   fprintf(stderr,"\nOptions requiring a customized view routine:\n");  
   fprintf(stderr," -1 shuffels nibbles of every second byte in bitmap\n");  
   fprintf(stderr," -2 shuffels bytes to pattern 0,2,1,3\n");
   fprintf(stderr," -3 shuffels bytes to pattern 0,2,4,6,1,3,5,7\n");     
   fprintf(stderr," -c pack 1000 color halfbytes in koala image to 500 bytes\n");     
   fprintf(stderr," -4 store value minus previous byte in bitmap\n");     
   fprintf(stderr," -5 store value XOR previous byte in bitmap\n");      
   fprintf(stderr," -8 store data columnwise\n");      
   fprintf(stderr," -d deinterleave color data\n");
   fprintf(stderr," -s structure define order of components in memory,\n"); 
   fprintf(stderr,"    default is BSCG (Bitmap, Screen, Colorram, backGround color), p=padding to next kB\n");   
                
   fprintf(stderr,"\nGeneral options, work with any file:\n");  
   fprintf(stderr," -r runlength encoding of file, first byte is escape byte\n");  
   
   fprintf(stderr,"\nBest options to achieve an image that can be compressed well are -8cz\n"); 
        
   exit(1);
}

int main(int argc, char ** argv) {
  int c;
	
  char *infilename=NULL;
  char *outfilename=NULL;
  int opt_nibble=0;
  int opt_bytes0213=0;  
  int opt_bytes02461357=0;
  int opt_XB=0; 
  int opt_XorB=0;
  int opt_colbytes=0;   
  int opt_rle=0;    
  int opt_columnwise=0;
  int opt_blankout=0;
  int opt_dint=0;  
  int opt_reorganizecolors=0;
  int opt_filestructure=0;  

  progname=argv[0];
  while ((c=getopt(argc,argv,"i:o:123458crdbzp:s:"))!=EOF)
  {
     switch(c)
     {
     case 'i':
     	if (infilename!=NULL) usage("Only one infile can be given.");
        infilename=optarg;
        break;      	
     case 'o':
     	if (outfilename!=NULL) usage("Only one outfile can be given.");     	
        outfilename=optarg;
        break;   
     case 'p':
     	if (opt_reorganizecolors==0) usage("Only possible with option -z.");
        palette=optarg;
        break;      
     case 's':
        fstructure=optarg;
        opt_filestructure++;
        break;          		           
     case '1':
        opt_nibble++;
        break;
     case '2':
        opt_bytes0213++;
        break;
     case '3':
        opt_bytes02461357++;
        break;        
     case '4':
        opt_XB++;
        break;
     case '5':
        opt_XorB++;
        break;
     case '8':
        opt_columnwise++;
        break;					        
     case 'c':
        opt_colbytes++;
        break;        
     case 'r':
        opt_rle++;
        break;     
     case 'b':
        opt_blankout++;
        break;   
     case 'z':
        opt_reorganizecolors++;
        break; 				  
     case 'd':
        opt_dint++;
        break;                     
     default:
        usage(NULL);
     }
  }
  
  if (optind-argc!=0) usage(NULL);
  
  /*infilename="uhrturm64.kla.prg";
  outfilename="test1.prg";
  opt_nibble++;
  */
  
  if (infilename==NULL) usage("No infile given!");
  if (outfilename==NULL) usage("No outfile given!");

  cbmfile_t filedata=load_file(infilename);

  if (opt_blankout) filedata=blankoutunusedcolors(filedata); 
  if (opt_reorganizecolors ) filedata=reorganizecolors(filedata); 
  if (opt_columnwise) filedata=columnwise(filedata);
  if (opt_nibble) filedata=shuffle_nibbles(filedata);
  if (opt_bytes0213) filedata=shuffle_bytes0213(filedata);
  if (opt_bytes02461357) filedata=shuffle_bytes02461357(filedata);  
  if (opt_XB==1)
    filedata=storediffs(filedata,8000);
  else if (opt_XB==2)
    filedata=storediffs(filedata,filedata.length);  
  if (opt_XorB==1)
    filedata=storexors(filedata,8000);
  else if (opt_XorB==2)
    filedata=storexors(filedata,filedata.length);    
 
  if (opt_colbytes) filedata=packcolbytes(filedata);   
  if (opt_dint) filedata=dintcolordata(filedata);     
  
  if (opt_filestructure) filedata=filestructure(filedata,fstructure);
  
//  if (opt_rle) filedata=RLE(filedata);    
  
  save_file(filedata,outfilename);
  
  return 0;
}


